#!/usr/bin/env python
'''
garage
Created by Seria at 03/01/2019 8:32 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-
from math import ceil
from ... import dock
from .architect import ConvG, ConvD, ResG, ResD, BN



class Discriminator(dock.Craft):
    # def __init__(self, in_shape, category_dim, code_dim, scope='DSC'):
    def __init__(self, in_shape, category_dim, code_dim, base_chs, norm_fn, attention, spec_norm, w_init, scope='DSC'):
        super(Discriminator, self).__init__(scope)
        H, W, C = in_shape
        self.H = H
        self.W = W
        self.C = C
        self.category_dim = category_dim
        self.code_dim = code_dim
        self.sftm = dock.Sftm()

        # self.backbone = ConvD(in_shape)
        # feat_dim = ceil(H/16) * ceil(W/16) * 128
        # self.cls = dock.Dense(feat_dim, 1)
        # if category_dim > 0:
        #     self.category = dock.Dense(feat_dim, category_dim)
        # if code_dim > 0:
        #     self.code = dock.Dense(feat_dim, code_dim)

        ############# FENCE ##############
        min_size = min(H, W)
        factor = {128: base_chs / 16}
        self.backbone = ResD(in_shape, base_chs, norm_fn, attention, spec_norm, w_init)
        feat_dim = int(H * W * factor[min_size])
        if spec_norm:
            self.cls = dock.SN(dock.Dense(feat_dim, 1))
            if category_dim > 0:
                self.category = dock.SN(dock.Dense(feat_dim, category_dim))
            if code_dim > 0:
                self.code = dock.SN(dock.Dense(feat_dim, code_dim))
        else:
            self.cls = dock.Dense(feat_dim, 1)
            if category_dim > 0:
                self.category = dock.Dense(feat_dim, category_dim)
            if code_dim > 0:
                self.code = dock.Dense(feat_dim, code_dim)

    def run(self, x):
        self['input'] = x
        x = self.backbone(x)

        self['cls'] = self.cls(x)
        y = [self['cls']]
        if self.category_dim > 0:
            self['category'] = self.sftm(self.category(x))
            y.append(self['category'])
        if self.code_dim > 0:
            self['code'] = self.code(x)
            y.append(self['code'])

        if len(y) > 1:
            return tuple(y)
        else:
            return y[0]



class InfoGAN(dock.Craft):
    # def __init__(self, in_shape, category_dim=0, code_dim=0, latent_dim=128, scope='INFOGAN'):
    def __init__(self, in_shape, category_dim=0, code_dim=0, latent_dim=128, base_chs=32, norm_fn=BN,
                 attention=False, spec_norm=False, w_init=dock.XavierNorm(), scope='INFOGAN'):
        super(InfoGAN, self).__init__(scope)
        # self.G = ConvG(in_shape, category_dim + code_dim + latent_dim)
        # self.D = Discriminator(in_shape, category_dim, code_dim)
        self.G = ResG(in_shape, category_dim + code_dim + latent_dim, base_chs, norm_fn, attention, spec_norm, w_init)
        self.D = Discriminator(in_shape, category_dim, code_dim, base_chs, norm_fn, attention, spec_norm, w_init)

    def run(self):
        pass


#!/usr/bin/env python
'''
component_tf
Created by Seria at 05/02/2019 1:41 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-
from functools import partial
import tensorflow as tf
import os
from ..toolkit.utility import recordConfig
from .component import Pod



class OffTheShelfTF(object):
    def __init__(self, **kwargs):
        super(OffTheShelfTF, self).__init__()

    def run(self, **kwargs):
        raise NotImplementedError

    def __call__(self, **kwargs):
        return self.run(**kwargs)

class ComponentTF(object):
    def __init__(self, channel_major=True, time_major=False):
        self.channel_major = channel_major
        self.time_major = time_major
        self.warehouse = ['CONV', 'TRANS_CONV', 'SEP_CONV',
                          'SIGMOID', 'TANH', 'SOFTMAX', 'RELU', 'RELU_LEAKY', 'RELU_EXP',
                          'MAX_POOL', 'AVG_POOL',
                          'DROPOUT', 'BATCH_NORM', 'EMBEDDING',
                          'RESHAPE', 'FLAT', 'SLICE',
                          'CLIP', 'DENSE',
                          'RESIZE',
                          'DUPLICATE', 'CONVERT',
                          'WEIGHT_DECAY', 'SIGM_XE', 'SFTM_XE', 'MSE', 'MAE',
                          'MOMENTUM', 'NESTEROV', 'ADAM']
        # Convolution
        self.CONV = self.conv
        self.TRANS_CONV = self.trans_conv
        self.SEP_CONV = 4
        # Activation
        self.SIGMOID = self.sigmoid
        self.TANH = self.tanh
        self.SOFTMAX = self.softmax
        self.RELU = self.relu
        self.LRELU = self.relu_leaky
        self.ERELU = 15
        # Pooling
        self.MAX_POOL = self.max_pool
        self.AVG_POOL = self.avg_pool
        # Distributing
        self.DROPOUT = self.dropout
        self.BATCH_NORM = self.batch_norm
        self.EMBEDDING = self.embedding
        # Reshape
        self.RESHAPE = self.reshape
        self.FLAT = self.flat
        self.SLICE = self.slice
        self.PAD = self.pad
        # Arithmetic
        self.CLIP = self.clip
        self.DENSE = self.dense
        self.ARGMAX = self.argmax
        # Image Manipulation
        self.RESIZE = self.resize
        # Copy or Rename
        self.DUPLICATE = self.duplicate
        self.CONVERT = self.convert
        # Loss
        self.WEIGHT_DECAY = self.weight_decay
        self.SIGM_XE = self.sigm_xentropy
        self.SFTM_XE = self.sftm_xentropy
        self.MSE = self.mse
        self.MAE = self.mae
        # Optimizer
        self.MOMENTUM = self.momentum
        self.NESTEROV = self.nesterov
        self.ADAM = self.adam
        # Metric
        self.ACC_CLS = self.accuracy_cls

    @staticmethod
    def doff(name):
        return os.path.basename(name)

    @staticmethod
    def cap(name, scope):
        return scope + ':' + name

    def remodel(self, name, comp, is_ready=False):
        name = name.upper()
        if name in self.warehouse:
            raise Exception('NEBULAE ERROR ⨷ %s is an existing component in warehouse.' % name)
        else:
            self.warehouse.append(name)
        if is_ready:
            def customizedComp(name):
                return Pod(partial(comp.component, name=name), comp.symbol, name, comp.message, comp.info)
        else:
            assert not isinstance(comp.component, list)
            def customizedComp(**kwargs):
                message = []
                for s, sym in enumerate(comp.symbol):
                    if sym in kwargs:
                        message.append(kwargs[sym])
                    else:
                        message.append(comp.message[s])
                info = {}
                for k, v in comp.info.items():
                    info[k] = kwargs.get(k, v)
                return Pod(partial(comp.component, **kwargs), comp.symbol, kwargs['name'], message, info)
        exec('self.%s = customizedComp' % name)

    def new(self, name, comp, *exec_args, out_shape):
        '''
        exec_args: (str,) input tensor names
        out_shape: output shape with batch size as -1
        '''
        name = name.upper()
        if name in self.warehouse:
            raise Exception('NEBULAE ERROR ⨷ %s is an existing component in warehouse.' % name)

        def newComp(**kwargs):
            message = []
            init_args = {}
            for k in kwargs:
                if k in exec_args:
                    message.append(kwargs[k])
                else:
                    init_args[k] = kwargs[k]
            return Pod(comp(**init_args), list(exec_args), kwargs['name'], message,
                       {'type': 'OTS', 'shape': out_shape})
        exec('self.%s = newComp' % name)

    # ------------------------------------ Convolution ------------------------------------ #

    def _createVar(self, name, shape, initializer, param, regularizer=None):
        init_err = Exception('NEBULAE ERROR ⨷ %s initializer is not defined or supported.' % initializer)
        # with tf.device('/gpu:0'):
        if initializer == 'xavier':
            if param=='normal':
                param = False
            elif param=='uniform':
                param = True
            else:
                param = False
            var = tf.get_variable(name, shape, initializer=tf.contrib.layers.xavier_initializer(uniform=param))
        elif initializer == 'uniform':
            if param is None:
                param = [-0.5, 0.5]
            var = tf.get_variable(name,
                                  initializer=tf.random.uniform(shape, minval=param[0], maxval=param[1]))
        elif initializer == 'normal':
            if param is None:
                param = [0, 0.1]
            var = tf.get_variable(name,
                                  initializer=tf.random.normal(shape, mean=param[0], stddev=param[1]))
        elif initializer == 'zero':
            var = tf.get_variable(name, shape, initializer=tf.zeros_initializer())
        elif initializer == 'one':
            var = tf.get_variable(name, shape, initializer=tf.ones_initializer())
        else:
            raise init_err

        if regularizer == 'l2':
            weight_decay = tf.nn.l2_loss(var, name=name+'/regularizer')
            tf.add_to_collection(tf.GraphKeys.REGULARIZATION_LOSSES, weight_decay)
        elif regularizer is None:
            pass
        else:
            raise Exception('NEBULAE ERROR ⨷ %s regularizer is not defined or supported.' % regularizer)

        return var

    def conv(self, **kwargs):
        message = kwargs.get('inputs', '_INPUTS')
        return Pod(partial(self._conv, **kwargs), ['inputs'], kwargs['name'], [message], {'type': 'CONV'})
    # TODO: implement group convolution in tf
    def _conv(self, name, inputs, out_chs, kernel, w_init='xavier', w_param=None, b_init=None, b_param=None,
                w_reg='l2', b_reg='l2', stride=(1, 1), dilation=(1, 1), group=1, auto_pad=True):
        if auto_pad:
            padding = 'SAME'
        else:
            padding = 'VALID'
        if self.channel_major:
            in_chs = int(inputs.get_shape()[1])
        else:
            in_chs = int(inputs.get_shape()[-1])
        in_out_chs = (in_chs, out_chs)
        w = self._createVar(name+'_w', kernel + in_out_chs, w_init, w_param, w_reg)
        dim = len(kernel)
        if dim == 1:
            conv_func = tf.nn.conv1d
            if self.channel_major:
                dform = 'NCW'
                stride = [1, 1] + list(stride)
                dilation = [1, 1] + list(dilation)
            else:
                dform = 'NWC'
                stride = [1] + list(stride) + [1]
                dilation = [1] + list(dilation) + [1]
        elif dim == 2:
            conv_func = tf.nn.conv2d
            if self.channel_major:
                dform = 'NCHW'
                stride = [1, 1] + list(stride)
                dilation = [1, 1] + list(dilation)
            else:
                dform = 'NHWC'
                stride = [1] + list(stride) + [1]
                dilation = [1] + list(dilation) + [1]
        elif dim == 3:
            conv_func = tf.nn.conv3d
            if self.channel_major:
                dform = 'NCDHW'
                stride = [1, 1] + list(stride)
                dilation = [1, 1] + list(dilation)
            else:
                dform = 'NDHWC'
                stride = [1] + list(stride) + [1]
                dilation = [1] + list(dilation) + [1]
        else:
            raise Exception('NEBULAE ERROR ⨷ %d-d convolution is not supported.' % dim)
        if b_init:
            b = self._createVar(name+'_b', [in_out_chs[-1]], b_init, b_param, b_reg)
            return tf.nn.bias_add(
                    conv_func(inputs, w, stride, padding, dilations=dilation, data_format=dform),
                    b, data_format=dform, name=name)
        else:
            return conv_func(inputs, w, stride, padding, dilations=dilation, data_format=dform, name=name)

    def trans_conv(self, **kwargs):
        message = kwargs.get('inputs', '_INPUT')
        return Pod(partial(self._trans_conv, **kwargs), ['inputs'], kwargs['name'], [message], {'type': 'TCONV'})
    def _trans_conv(self, name, inputs, output, kernel, w_init='xavier', w_param=None, b_init=None, b_param=None,
              w_reg='l2', b_reg='l2', stride=(1, 1), dilation=(1, 1), group=1, auto_pad=True):
        if auto_pad:
            padding = 'SAME'
        else:
            padding = 'VALID'
        if self.channel_major:
            in_chs = int(inputs.get_shape()[1])
            out_chs = output[1]
        else:
            in_chs = int(inputs.get_shape()[-1])
            out_chs = output[-1]
        in_out_chs = (out_chs, in_chs)
        w = self._createVar(name + '_w', kernel + in_out_chs, w_init, w_param, w_reg)
        dim = len(kernel)
        if dim == 1:
            conv_func = tf.nn.conv1d_transpose
            if self.channel_major:
                dform = 'NCW'
                stride = [1, 1] + list(stride)
                dilation = [1, 1] + list(dilation)
            else:
                dform = 'NWC'
                stride = [1] + list(stride) + [1]
                dilation = [1] + list(dilation) + [1]
        elif dim == 2:
            conv_func = tf.nn.conv2d_transpose
            if self.channel_major:
                dform = 'NCHW'
                stride = [1, 1] + list(stride)
                dilation = [1, 1] + list(dilation)
            else:
                dform = 'NHWC'
                stride = [1] + list(stride) + [1]
                dilation = [1] + list(dilation) + [1]
        elif dim == 3:
            conv_func = tf.nn.conv3d_transpose
            if self.channel_major:
                dform = 'NCDHW'
                stride = [1, 1] + list(stride)
                dilation = [1, 1] + list(dilation)
            else:
                dform = 'NDHWC'
                stride = [1] + list(stride) + [1]
                dilation = [1] + list(dilation) + [1]
        else:
            raise Exception('NEBULAE ERROR ⨷ %d-d convolution is not supported.' % dim)
        if b_init:
            b = self._createVar(name + '_b', [in_out_chs[-2]], b_init, b_param, b_reg)
            return tf.nn.bias_add(
                    conv_func(inputs, w, tf.constant(output), stride, padding, dform, dilations=dilation),
                    b, data_format=dform, name=name)
        else:
            return conv_func(inputs, w, tf.constant(output), stride, padding, dform, dilations=dilation, name=name)

    # ------------------------------------ Activation ------------------------------------ #

    def sigmoid(self, **kwargs):
        message = kwargs.get('inputs', '_INPUTS')
        return Pod(partial(self._sigmoid, **kwargs), ['inputs'], kwargs['name'], [message], {'type': 'ACTIVATION'})
    def _sigmoid(self, name, inputs):
        return tf.nn.sigmoid(inputs, name)

    def tanh(self, **kwargs):
        message = kwargs.get('inputs', '_INPUTS')
        return Pod(partial(self._tanh, **kwargs), ['inputs'], kwargs['name'], [message], {'type': 'ACTIVATION'})
    def _tanh(self, name, inputs):
        return tf.nn.tanh(inputs, name)

    def softmax(self, **kwargs):
        message = kwargs.get('inputs', '_INPUTS')
        return Pod(partial(self._tanh, **kwargs), ['inputs'], kwargs['name'], [message], {'type': 'ACTIVATION'})
    def _softmax(self, name, inputs):
        return tf.nn.softmax(inputs, name)

    def relu(self, **kwargs):
        message = kwargs.get('inputs', '_INPUTS')
        return Pod(partial(self._relu, **kwargs), ['inputs'], kwargs['name'], [message], {'type': 'ACTIVATION'})
    def _relu(self, name, inputs):
        return tf.nn.relu(inputs, name)

    def relu_leaky(self, **kwargs):
        message = kwargs.get('inputs', '_INPUTS')
        return Pod(partial(self._relu_leaky, **kwargs), ['inputs'], kwargs['name'], [message], {'type': 'ACTIVATION'})
    def _relu_leaky(self, name, inputs, alpha=0.2):
        return tf.nn.leaky_relu(inputs, alpha, name)

    # ------------------------------------ Distributing ------------------------------------ #

    def dropout(self, **kwargs):
        message = []
        message.append(kwargs.get('inputs', '_INPUTS'))
        message.append(kwargs.get('is_train', '_IS_TRAIN'))
        return Pod(partial(self._dropout, **kwargs), ['inputs', 'is_train'], kwargs['name'], message, {'type': 'DROP'})
    def _dropout(self, name, inputs, p_drop, is_train):
        if is_train:
            p_keep = 1-p_drop
        else:
            p_keep = 1
        return tf.nn.dropout(inputs, p_keep)

    def batch_norm(self, **kwargs):
        message = []
        message.append(kwargs.get('inputs', '_INPUTS'))
        message.append(kwargs.get('is_train', '_IS_TRAIN'))
        return Pod(partial(self._batch_norm, **kwargs), ['inputs', 'is_train'], kwargs['name'], message, {'type': 'BN'})
    def _batch_norm(self, name, inputs, is_train, mmnt=0.9, beta=False, gamma=False):
        if self.channel_major:
            axis = 1
        else:
            axis = -1
        return tf.layers.batch_normalization(inputs, training=is_train, momentum=mmnt, center=beta, scale=gamma,
                                             axis=axis, epsilon=1e-5, name=name)

    def embedding(self, **kwargs):
        message = kwargs.get('inputs', '_INPUTS')
        return Pod(partial(self._embedding, **kwargs), ['inputs'], kwargs['name'], [message], {'type': 'EMBED'})
    def _embedding(self, name, inputs, vocabulary, vec_dims, w_init='xavier', w_param=None):
        embd_vec = self._createVar(name+'_vec', [vocabulary, vec_dims], w_init, w_param)
        return tf.nn.embedding_lookup(embd_vec, inputs, name=name)

    # ------------------------------------ Pooling ------------------------------------ #

    def max_pool(self, **kwargs):
        message = kwargs.get('inputs', '_INPUTS')
        return Pod(partial(self._max_pool, **kwargs), ['inputs'], kwargs['name'], [message], {'type': 'POOL'})
    def _max_pool(self, name, inputs, kernel=(2, 2), stride=(2, 2), auto_pad=True, if_global=False):
        if auto_pad:
            padding = 'SAME'
        else:
            padding = 'VALID'
        dim = len(kernel)
        if dim == 2:
            pool_func = tf.nn.max_pool2d
            if self.channel_major:
                dform = 'NCHW'
                kernel = [1, 1] + list(kernel)
                stride = [1, 1] + list(stride)
            else:
                dform = 'NHWC'
                kernel = [1] + list(kernel) + [1]
                stride = [1] + list(stride) + [1]
        elif dim == 3:
            pool_func = tf.nn.max_pool3d
            if self.channel_major:
                dform = 'NCDHW'
                kernel = [1, 1] + list(kernel)
                stride = [1, 1] + list(stride)
            else:
                dform = 'NDHWC'
                kernel = [1] + list(kernel) + [1]
                stride = [1] + list(stride) + [1]
        else:
            raise Exception('NEBULAE ERROR ⨷ %d-d pooling is not supported.' % dim)
        if if_global:
            in_shape = inputs.get_shape()
            if self.channel_major:
                kernel = [1, 1] + [int(in_shape[-d] for d in range(dim, 0, -1))]
            else:
                kernel = [1] + [int(in_shape[-d] for d in range(dim+1, 1, -1))] + [1]
        return pool_func(inputs, kernel, stride, padding, data_format=dform, name=name)

    def avg_pool(self, **kwargs):
        message = kwargs.get('inputs', '_INPUTS')
        return Pod(partial(self._avg_pool, **kwargs), ['inputs'], kwargs['name'], [message], {'type': 'POOL'})
    def _avg_pool(self, name, inputs, kernel=(2, 2), stride=(2, 2), auto_pad=True, if_global=False):
        if auto_pad:
            padding = 'SAME'
        else:
            padding = 'VALID'
        dim = len(kernel)
        if dim == 2:
            pool_func = tf.nn.avg_pool2d
            if self.channel_major:
                dform = 'NCHW'
                kernel = [1, 1] + list(kernel)
                stride = [1, 1] + list(stride)
            else:
                dform = 'NHWC'
                kernel = [1] + list(kernel) + [1]
                stride = [1] + list(stride) + [1]
        elif dim == 3:
            pool_func = tf.nn.avg_pool3d
            if self.channel_major:
                dform = 'NCDHW'
                kernel = [1, 1] + list(kernel)
                stride = [1, 1] + list(stride)
            else:
                dform = 'NDHWC'
                kernel = [1] + list(kernel) + [1]
                stride = [1] + list(stride) + [1]
        else:
            raise Exception('NEBULAE ERROR ⨷ %d-d pooling is not supported.' % dim)
        if if_global:
            in_shape = inputs.get_shape()
            if self.channel_major:
                kernel = [1, 1] + [int(in_shape[-d] for d in range(dim, 0, -1))]
            else:
                kernel = [1] + [int(in_shape[-d] for d in range(dim + 1, 1, -1))] + [1]
        return pool_func(inputs, kernel, stride, padding, data_format=dform, name=name)

    # ------------------------------------ Reshape ------------------------------------ #

    def reshape(self, **kwargs):
        message = kwargs.get('inputs', '_INPUTS')
        return Pod(partial(self._reshape, **kwargs), ['inputs'], kwargs['name'], [message], {'type': 'RESHAPE'})
    def _reshape(self, name, inputs, shape):
        return tf.reshape(inputs, shape, name)

    def flat(self, **kwargs):
        message = kwargs.get('inputs', '_INPUTS')
        return Pod(partial(self._flat, **kwargs), ['inputs'], kwargs['name'], [message], {'type': 'FLAT'})
    def _flat(self, name, inputs):
        return tf.layers.flatten(inputs, name=name)

    def slice(self, **kwargs):
        message = kwargs.get('inputs', '_INPUTS')
        return Pod(partial(self._slice, **kwargs), ['inputs'], kwargs['name'], [message], {'type': 'SLICE'})
    def _slice(self, name, inputs, indices):
        for d in range(len(indices)):
            output = inputs[indices[d][0]:indices[d][1]]
        return tf.identity(output, name)

    def pad(self, **kwargs):
        message = kwargs.get('inputs', '_INPUTS')
        return Pod(partial(self._pad, **kwargs), ['inputs'], kwargs['name'], [message], {'type': 'PAD'})
    def _pad(self, name, inputs, margin, fill_in=0):
        return tf.pad(inputs, margin, constant_values=fill_in, name=name)

    # -------------------------------------- Arithmetic -------------------------------------- #

    def clip(self, **kwargs):
        message = kwargs.get('inputs', '_INPUTS')
        return Pod(partial(self._clip, **kwargs), ['inputs'], kwargs['name'], [message], {'type': 'CLIP'})
    def _clip(self, name, inputs, min_max_vals):
        return tf.clip_by_value(inputs, min_max_vals[0], min_max_vals[1], name)

    def dense(self, **kwargs):
        message = kwargs.get('inputs', '_INPUTS')
        return Pod(partial(self._dense, **kwargs), ['inputs'], kwargs['name'], [message], {'type': 'DENSE'})
    def _dense(self, name, inputs, out_chs, axis=-1, w_init='xavier', w_param=None,
               b_init='zero', b_param=None, w_reg='l2', b_reg='l2'):
        size = inputs.get_shape().as_list()
        in_chs = size[axis]
        in_out_chs = [in_chs, out_chs]
        if axis == 0:
            raise Exception('NEBULAE ERROR ⨷ you cannot apply dense layer along batch axis.')
        elif axis == -1:
            w = self._createVar(name + '_w', in_out_chs, w_init, w_param, w_reg)
            if b_init:
                b = self._createVar(name + '_b', [in_out_chs[-1]], b_init, b_param, b_reg)
                return tf.nn.bias_add(tf.matmul(inputs, w), b, name=name)
            else:
                return tf.matmul(inputs, w, name=name)
        else:
            channel = (in_chs,)
            del size[axis]
            size[0] = -1
            in_shape = tuple(size) + channel
            out_shape = list(inputs[0])
            out_shape[0] = -1
            out_shape[axis] = out_chs
            out_shape = tuple(out_shape)
            w = self._createVar(name + '_w', in_out_chs, w_init, w_param, w_reg)
            if b_init:
                b = self._createVar(name + '_b', [in_out_chs[-1]], b_init, b_param, b_reg)
                return tf.reshape(tf.nn.bias_add(tf.matmul(tf.reshape(inputs, in_shape), w), b), out_shape, name=name)
            else:
                return tf.reshape(tf.matmul(tf.reshape(inputs, in_shape), w), out_shape, name=name)

    def argmax(self, **kwargs):
        message = kwargs.get('inputs', '_INPUTS')
        return Pod(partial(self._argmax, **kwargs), ['inputs'], kwargs['name'], [message], {'type': 'ARG'})
    def _argmax(self, name, inputs, axis):
        return tf.argmax(inputs, axis, name=name)

    # ------------------------------------ Image Manipulation ------------------------------------ #

    def resize(self, **kwargs):
        message = kwargs.get('inputs', '_INPUTS')
        return Pod(partial(self._resize, **kwargs), ['inputs'], kwargs['name'], [message], {'type': 'RESHAPE'})
    def _resize(self, name, inputs, size, method='bilinear'):
        if method == 'bilinear':
            return tf.image.resize_bilinear(inputs, size, name=name)
        elif method == 'bicubic':
            return tf.image.resize_bicubic(inputs, size, name=name)
        elif method == 'crop':
            return tf.image.resize_image_with_crop_or_pad(inputs, size, name=name)
        else:
            raise KeyError('NEBULAE ERROR ⨷ %s is not a legal resize method.' % method)


    # ------------------------------------ Redefine or Rename ------------------------------------ #

    def duplicate(self, **kwargs):
        message = kwargs.get('inputs', '_INPUTS')
        return Pod(partial(self._duplicate, **kwargs), ['inputs'], kwargs['name'], [message], {'type': 'DUP'})
    def _duplicate(self, name, inputs):
        return tf.identity(inputs, name)

    def convert(self, **kwargs):
        message = kwargs.get('inputs', '_INPUTS')
        return Pod(partial(self._convert, **kwargs), ['inputs'], kwargs['name'], [message], {'type': 'CAST'})
    def _convert(self, name, inputs, dtype, trainable=False):
        '''
        convert data type or convert list/numpy array to tensor
        :param name:
        :param inputs: input tensor / list / numpy array
        :param dtype:
        :param trainable: if tensor is trainable
        :return: tensor
        '''
        if isinstance(inputs, (tf.Tensor, tf.SparseTensor, tf.Variable)):
            return tf.cast(inputs, tf.as_dtype(dtype), name=name)
        else:
            return tf.Variable(inputs, trainable=trainable, name=name)

    # ------------------------------------ Loss ------------------------------------ #

    def weight_decay(self, **kwargs):
        # message = kwargs.get('penalty', '.penalty')
        return Pod(partial(self._weight_decay, **kwargs), [], kwargs['name'], [], {'type': 'LOSS'})
    def _weight_decay(self, name, penalty, decay_scope=None):
        if decay_scope is None:
            return tf.multiply(tf.add_n(tf.get_collection(tf.GraphKeys.REGULARIZATION_LOSSES)),
                               penalty, name=name)
        else:
            return tf.multiply(tf.add_n(tf.get_collection(tf.GraphKeys.REGULARIZATION_LOSSES, scope=decay_scope)),
                               penalty, name=name)

    def sigm_xentropy(self, **kwargs):
        message = []
        message.append(kwargs.get('inputs', '_INPUTS'))
        message.append(kwargs.get('label', '_LABEL'))
        return Pod(partial(self._sigm_xentropy, **kwargs), ['inputs', 'label'], kwargs['name'], message, {'type': 'LOSS'})
    def _sigm_xentropy(self, name, inputs, label):
        labels = tf.cast(label, tf.float32)
        return tf.reduce_mean(tf.nn.sigmoid_cross_entropy_with_logits(labels=label, logits=inputs, name=name))

    def sftm_xentropy(self, **kwargs):
        message = []
        message.append(kwargs.get('inputs', '_INPUTS'))
        message.append(kwargs.get('label', '_LABEL'))
        return Pod(partial(self._sftm_xentropy, **kwargs), ['inputs', 'label'], kwargs['name'], message, {'type': 'LOSS'})
    def _sftm_xentropy(self, name, inputs, label, is_one_hot):
        if is_one_hot:
            labels = tf.cast(label, tf.float32)
            return tf.reduce_mean(tf.nn.softmax_cross_entropy_with_logits(labels=labels, logits=inputs, name=name))
        else:
            return tf.reduce_mean(tf.nn.sparse_softmax_cross_entropy_with_logits(
                                        labels=label, logits=inputs, name=name))

    def mse(self, **kwargs):
        message = []
        message.append(kwargs.get('inputs', '_INPUTS'))
        message.append(kwargs.get('label', '_LABEL'))
        return Pod(partial(self._mse, **kwargs), ['inputs', 'label'], kwargs['name'], message, {'type': 'LOSS'})
    def _mse(self, name, inputs, label):
        return tf.reduce_mean(tf.squared_difference(inputs, label), name=name)

    def mae(self, **kwargs):
        message = []
        message.append(kwargs.get('inputs', '_INPUTS'))
        message.append(kwargs.get('label', '_LABEL'))
        return Pod(partial(self._mae, **kwargs), ['inputs', 'label'], kwargs['name'], message, {'type': 'LOSS'})
    def _mae(self, name, inputs, label):
        return tf.reduce_mean(tf.abs(inputs-label), name=name)

    # ------------------------------------ Optimizer ------------------------------------ #

    def _lrStrategy(self, mileage, lr, lr_decay, miles, param):
        if lr_decay == 'step':
            return tf.train.exponential_decay(lr, mileage, miles, decay_rate=param, staircase=True)
        elif lr_decay == 'poly':
            return tf.train.polynomial_decay(lr, mileage, miles, power=param, cycle=False)
        elif lr_decay == 'cosine':
            return tf.train.cosine_decay(lr, mileage, miles, alpha=lr*0.001)
        elif lr_decay == 'exp':
            return tf.train.exponential_decay(lr, mileage, miles, decay_rate=param, staircase=False)
        elif lr_decay == 'wavy':
            return tf.train.cosine_decay_restarts(lr, mileage, miles, t_mul=1.0, m_mul=param, alpha=lr*0.001)
        else:
            raise KeyError('NEBULAE ERROR ⨷ %s decay is not supported or defined.' % lr_decay)

    def _computeGrad(self, optz, cost, update_scope, ignore_name, grad_limit):
        update_ops = tf.get_collection(tf.GraphKeys.UPDATE_OPS)
        with tf.control_dependencies(update_ops):
            if update_scope is None:
                tr_var = tf.trainable_variables()
            else:
                tr_var = tf.trainable_variables(scope=update_scope)
            if not ignore_name is None:
                update_var = []
                for tv in tr_var:
                    flag_ignore = False
                    for ig in ignore_name:
                        if ig in tv.op.name:
                            flag_ignore = True
                            break
                    if not flag_ignore:
                        update_var.append(tv)
            else:
                update_var = tr_var
            grad_var_pairs = optz.compute_gradients(cost, var_list=update_var)

            if grad_limit is None:
                return grad_var_pairs
            else:
                return [(tf.clip_by_value(grad, -grad_limit, grad_limit), var) for grad, var in grad_var_pairs]

    def momentum(self, **kwargs):
        message = kwargs.get('inputs', '_INPUTS')
        return Pod(partial(self._momentum, **kwargs), ['inputs', 'media'], kwargs['name'], [message, '_MEDIA'], {'type': 'OPTZ'})
    def _momentum(self, name, inputs, lr, mmnt=0.9, update_scope=None, ignore_name=None,
                  lr_decay=None, lr_miles=None, decay_param=None, grad_limit=None, media=None):
        recordConfig(os.path.join(os.getcwd(), 'nebulae_temp_config.json'),
                     {'lr': lr, 'mmnt': mmnt, 'update_scope': update_scope, 'ignore_name': ignore_name,
                      'lr_decay': lr_decay, 'lr_miles': lr_miles, 'decay_param': decay_param, 'grad_limit': grad_limit},
                     overwrite=False)
        if isinstance(lr_decay, str):
            lr = self._lrStrategy(media, lr, lr_decay, lr_miles, decay_param)
        optz = tf.train.MomentumOptimizer(lr, mmnt)
        grad = self._computeGrad(optz, inputs, update_scope, ignore_name, grad_limit)
        train_op = optz.apply_gradients(grad, media, name=name)
        return train_op

    def nesterov(self, **kwargs):
        message = kwargs.get('inputs', '_INPUTS')
        return Pod(partial(self._nesterov, **kwargs), ['inputs', 'media'], kwargs['name'], [message, '_MEDIA'], {'type': 'OPTZ'})
    def _nesterov(self, name, inputs, lr, mmnt=0.9, update_scope=None, ignore_name=None,
                  lr_decay=None, lr_miles=None, decay_param=None, grad_limit=None, media=None):
        recordConfig(os.path.join(os.getcwd(), 'nebulae_temp_config.json'),
                     {'lr': lr, 'mmnt': mmnt, 'update_scope': update_scope, 'ignore_name': ignore_name,
                      'lr_decay': lr_decay, 'lr_miles': lr_miles, 'decay_param': decay_param, 'grad_limit': grad_limit},
                     overwrite=False)
        if isinstance(lr_decay, str):
            lr = self._lrStrategy(media, lr, lr_decay, lr_miles, decay_param)
        optz = tf.train.MomentumOptimizer(lr, mmnt, use_nesterov=True)
        grad = self._computeGrad(optz, inputs, update_scope, ignore_name, grad_limit)
        train_op = optz.apply_gradients(grad, media, name=name)
        return train_op

    def adam(self, **kwargs):
        message = kwargs.get('inputs', '_INPUTS')
        return Pod(partial(self._adam, **kwargs), ['inputs', 'media'], kwargs['name'], [message, '_MEDIA'], {'type': 'OPTZ'})
    def _adam(self, name, inputs, lr, mmnt1=0.9, mmnt2=0.999, update_scope=None, ignore_name=None,
                  lr_decay=None, lr_miles=None, decay_param=None, grad_limit=None, media=None):
        recordConfig(os.path.join(os.getcwd(), 'nebulae_temp_config.json'),
                     {'lr': lr, 'mmnt1': mmnt1, 'mmnt2': mmnt2, 'update_scope': update_scope,
                      'ignore_name': ignore_name, 'lr_decay': lr_decay, 'lr_miles': lr_miles,
                      'decay_param': decay_param, 'grad_limit': grad_limit},
                     overwrite=False)
        if isinstance(lr_decay, str):
            lr = self._lrStrategy(media, lr, lr_decay, lr_miles, decay_param)
        optz = tf.train.AdamOptimizer(lr, beta1=mmnt1, beta2=mmnt2)
        grad = self._computeGrad(optz, inputs, update_scope, ignore_name, grad_limit)
        train_op = optz.apply_gradients(grad, media, name=name)
        return train_op

    # ------------------------------------ Metric ------------------------------------ #

    def accuracy_cls(self, **kwargs):
        message = []
        message.append(kwargs.get('inputs', '_INPUTS'))
        message.append(kwargs.get('label', '_LABEL'))
        return Pod(partial(self._accuracy_cls, **kwargs), ['inputs', 'label'], kwargs['name'], message, {'type': 'ACC'})
    def _accuracy_cls(self, name, inputs, label, multi_class, is_one_hot=False):
        if multi_class:
            assert not is_one_hot
            pred = tf.cast(tf.round(inputs), label.dtype)
            correct = tf.reduce_all(tf.equal(pred, label), axis=-1)
            return tf.reduce_mean(tf.cast(correct, tf.float32), name=name)
        else:
            if is_one_hot:
                label = tf.argmax(label, axis=-1)
            pred = tf.argmax(inputs, axis=-1)
            correct = tf.equal(pred, label)
            return tf.reduce_mean(tf.cast(correct, tf.float32), name=name)


#!/usr/bin/env python
'''
component_tf
Created by Seria at 05/02/2019 1:41 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-
from functools import partial
from mxnet.gluon import nn, rnn, loss, Trainer, HybridBlock, ParameterDict
from mxnet import nd, init, autograd, lr_scheduler
from math import ceil, cos, pi
import os
from ..toolkit.utility import recordConfig
from .component import Pod



class OffTheShelfMX(HybridBlock):
    def __init__(self, **kwargs):
        prefix = kwargs.get('prefix', None)
        super(OffTheShelfMX, self).__init__(prefix=prefix)

    def run(self, *args):
        raise NotImplementedError

    def hybrid_forward(self, F, *args):
        return self.run(*args)

class CosineAnnealingSchedule():
    def __init__(self, min_lr, max_lr, cycle_length):
        """
        min_lr: lower bound for learning rate (float)
        max_lr: upper bound for learning rate (float)
        cycle_length: iterations between start and finish (int)
        """
        self.min_lr = min_lr
        self.max_lr = max_lr
        self.cycle_length = cycle_length

    def __call__(self, iteration):
        if iteration <= self.cycle_length:
            unit_cycle = (1 + cos(iteration * pi / self.cycle_length)) / 2
            adjusted_cycle = (unit_cycle * (self.max_lr - self.min_lr)) + self.min_lr
            return adjusted_cycle
        else:
            return self.min_lr

class CyclicalSchedule():
    def __init__(self, schedule_class, cycle_length, cycle_length_decay=1, cycle_magnitude_decay=1, **kwargs):
        """
        schedule_class: class of schedule, expected to take `cycle_length` argument
        cycle_length: iterations used for initial cycle (int)
        cycle_length_decay: factor multiplied to cycle_length each cycle (float)
        cycle_magnitude_decay: factor multiplied learning rate magnitudes each cycle (float)
        kwargs: passed to the schedule_class
        """
        self.schedule_class = schedule_class
        self.length = cycle_length
        self.length_decay = cycle_length_decay
        self.magnitude_decay = cycle_magnitude_decay
        self.kwargs = kwargs

    def __call__(self, iteration):
        cycle_idx = 0
        cycle_length = self.length
        idx = self.length
        while idx <= iteration:
            cycle_length = ceil(cycle_length * self.length_decay)
            cycle_idx += 1
            idx += cycle_length
        cycle_offset = iteration - idx + cycle_length

        schedule = self.schedule_class(cycle_length=cycle_length, **self.kwargs)
        return schedule(cycle_offset) * self.magnitude_decay**cycle_idx

class MXsigmoid(HybridBlock):
    def __init__(self, **kwargs):
        super(MXsigmoid, self).__init__(**kwargs)

    def hybrid_forward(self, F, input):
        return F.sigmoid(input)

class MXtanh(HybridBlock):
    def __init__(self, **kwargs):
        super(MXtanh, self).__init__(**kwargs)

    def hybrid_forward(self, F, input):
        return F.tanh(input)

class MXsoftmax(HybridBlock):
    def __init__(self, **kwargs):
        super(MXsoftmax, self).__init__(**kwargs)

    def hybrid_forward(self, F, input):
        return F.softmax(input)

class MXrelu(HybridBlock):
    def __init__(self, **kwargs):
        super(MXrelu, self).__init__(**kwargs)

    def hybrid_forward(self, F, input):
        return F.relu(input)

class MXreshape(HybridBlock):
    def __init__(self, *shape, **kwargs):
        super(MXreshape, self).__init__(**kwargs)
        self.shape = shape

    def hybrid_forward(self, F, input):
        return input.reshape(*self.shape)

class MXdense(HybridBlock):
    def __init__(self, out_chs, use_bias, w_init, b_init, in_shape, out_shape, **kwargs):
        super(MXdense, self).__init__(**kwargs)
        self.in_shape = in_shape
        self.out_shape = out_shape
        self.fc = nn.Dense(out_chs, use_bias=use_bias, flatten=False, weight_initializer=w_init, bias_initializer=b_init)

    def hybrid_forward(self, F, input):
        if self.in_shape is None:
            output = self.fc(input)
        else:
            output = input.reshape(self.in_shape)
            output = self.fc(output)
            output = output.reshape(self.out_shape)
        return output

class Duplicate(HybridBlock):
    def __init__(self, **kwargs):
        super(Duplicate, self).__init__(**kwargs)

    def hybrid_forward(self, F, input):
        return input

class MXwd(HybridBlock):
    def __init__(self, **kwargs):
        super(MXwd, self).__init__(**kwargs)

    def hybrid_forward(self, F, penalty):
        return F.zeros((1))

class MXmse(HybridBlock):
    def __init__(self, **kwargs):
        super(MXmse, self).__init__(**kwargs)

    def hybrid_forward(self, F, input, label):
        return (input-label)**2

class MXmae(HybridBlock):
    def __init__(self, **kwargs):
        super(MXmae, self).__init__(**kwargs)

    def hybrid_forward(self, F, input, label):
        return F.abs(input-label)

class MXsigmxe(HybridBlock):
    def __init__(self, **kwargs):
        super(MXsigmxe, self).__init__(**kwargs)
        self.cost = loss.SigmoidBCELoss()

    def hybrid_forward(self, F, input, label):
        return F.mean(self.cost(input, label))

class MXsftmxe(HybridBlock):
    def __init__(self, is_one_hot, **kwargs):
        super(MXsftmxe, self).__init__(**kwargs)
        self.cost = loss.SoftmaxCELoss(sparse_label=not is_one_hot)

    def hybrid_forward(self, F, input, label):
        return F.mean(self.cost(input, label))

class MX(HybridBlock):
    def __init__(self, **kwargs):
        super(MX, self).__init__(**kwargs)

    def hybrid_forward(self, F, input):
        return nd.sigmoid(input)

class AccuracyCLS(HybridBlock):
    def __init__(self, multi_class, is_one_hot, **kwargs):
        super(AccuracyCLS, self).__init__(**kwargs)
        if multi_class:
            assert not is_one_hot
        self.mulcls = multi_class
        self.onehot = is_one_hot

    def hybrid_forward(self, F, input, label):
        if self.mulcls:
            pred = F.round(input)
            correct = F.mean(pred == label, axis=-1)
            return F.mean(correct == 1)
        else:
            if self.onehot:
                label = F.argmax(label, axis=-1)
            pred = F.argmax(input, axis=-1)
            return F.mean(pred == label)

class ComponentMX(object):
    def __init__(self, channel_major=True, time_major=False):
        self.channel_major = channel_major
        self.time_major = time_major
        self.warehouse = ['CONV', 'TRANS_CONV', 'SEP_CONV',
                          'SIGMOID', 'TANH', 'SOFTMAX', 'RELU', 'RELU_LEAKY', 'RELU_EXP',
                          'MAX_POOL', 'AVG_POOL',
                          'DROPOUT', 'BATCH_NORM', 'EMBEDDING',
                          'RESHAPE', 'FLAT', 'SLICE',
                          'CLIP', 'DENSE',
                          'RESIZE',
                          'DUPLICATE', 'CONVERT',
                          'WEIGHT_DECAY', 'SIGM_XE', 'SFTM_XE', 'MSE', 'MAE',
                          'MOMENTUM', 'NESTEROV', 'ADAM']
        # Convolution
        self.CONV = self.conv
        self.TRANS_CONV = self.trans_conv
        self.SEP_CONV = 4
        # Activation
        self.SIGMOID = self.sigmoid
        self.TANH = self.tanh
        self.SOFTMAX = self.softmax
        self.RELU = self.relu
        self.LRELU = self.relu_leaky
        # self.ERELU = self.relu_exp
        # Pooling
        self.MAX_POOL = self.max_pool
        self.AVG_POOL = self.avg_pool
        # Distributing
        self.DROPOUT = self.dropout
        self.BATCH_NORM = self.batch_norm
        # self.EMBEDDING = self.embedding
        # Reshape
        self.RESHAPE = self.reshape
        self.FLAT = self.flat
        # self.SLICE = self.slice
        # self.PAD = self.pad
        # Arithmetic
        # self.CLIP = self.clip
        self.DENSE = self.dense
        # self.ARGMAX = self.argmax
        # Image Manipulation
        # self.RESIZE = self.resize
        # Copy or Rename
        self.DUPLICATE = self.duplicate
        # self.CONVERT = self.convert
        # Wrapped Net
        # self.RNN = self.rnn
        # self.LSTM = self.lstm
        # self.BILSTM = self.bilstm
        # Loss
        self.WEIGHT_DECAY = self.weight_decay
        self.SIGM_XE = self.sigm_xentropy
        self.SFTM_XE = self.sftm_xentropy
        self.MSE = self.mse
        self.MAE = self.mae
        # Optimizer
        self.MOMENTUM = self.momentum
        self.NESTEROV = self.nesterov
        self.ADAM = self.adam
        # Metric
        self.ACC_CLS = self.accuracy_cls

    @staticmethod
    def doff(name):
        return os.path.basename(name)

    @staticmethod
    def cap(name, scope):
        return scope + ':' + name

    def remodel(self, name, comp, is_ready=False):
        name = name.upper()
        if name in self.warehouse:
            raise Exception('NEBULAE ERROR ⨷ %s is an existing component in warehouse.' % name)
        else:
            self.warehouse.append(name)
        if is_ready:
            def customizedComp(name):
                return Pod(partial(comp.component, name=name), comp.symbol, name, comp.message, comp.info)
        else:
            assert not isinstance(comp.component, list)
            def customizedComp(**kwargs):
                message = []
                # check input so as to update message
                for s, sym in enumerate(comp.symbol):
                    if sym in kwargs:
                        message.append(kwargs[sym])
                    else:
                        message.append(comp.message[s])
                info = {}
                for k, v in comp.info.items():
                    info[k] = kwargs.get(k, v)
                return Pod(partial(comp.component, **kwargs), comp.symbol, kwargs['name'], message, info)
        exec('self.%s = customizedComp' % name)

    def new(self, name, comp, *exec_args, out_shape):
        '''
        exec_args: (str,) input tensor names
        out_shape: output shape with batch size as -1
        '''
        name = name.upper()
        if name in self.warehouse:
            raise Exception('NEBULAE ERROR ⨷ %s is an existing component in warehouse.' % name)

        def newComp(**kwargs):
            message = []
            init_args = {}
            for k in kwargs:
                if k in exec_args:
                    message.append(kwargs[k])
                else:
                    init_args[k] = kwargs[k]
            return Pod(partial(_newComp, **init_args), list(exec_args), kwargs['name'], message,
                       {'type': 'OTS', 'shape': out_shape})
        def _newComp(**init_args):
            return comp(**init_args)
        exec('self.%s = newComp' % name)

    # ------------------------------------ Convolution ------------------------------------ #

    def _getInitializer(self, initializer, param, regularizer):
        init_err = Exception('NEBULAE ERROR ⨷ %s initializer is not defined or supported.' % initializer)
        if initializer == 'xavier':
            if param is None:
                param = 'gaussian'
            else:
                if param=='normal':
                    param = 'gaussian'
                elif param!='uniform':
                    raise Exception('NEBULAE ERROR ⨷ %s-%s initializer is not defined or supported.' % (initializer, param))
            var = init.Xavier(param)
        elif initializer == 'uniform':
            if param is None:
                param = 0.5
            else:
                if param[0] != 0:
                    raise Exception('NEBULAE ERROR ⨷ the uniform distribution must be symmetric about y-axis in MXNet core.')
                param = param[1]
            var = init.Uniform(param)
        elif initializer == 'normal':
            if param is None:
                param = 0.1
            else:
                if param[0] != 0:
                    raise Exception('NEBULAE ERROR ⨷ the normal distribution must be zero-mean in MXNet core.')
                param = param[1]
            var = init.Normal(param)
        elif initializer == 'zero':
            var = init.Zero()
        elif initializer == 'one':
            var = init.One()
        else:
            raise init_err

        if regularizer == 'l2':
            pass
        else:
            raise Exception('NEBULAE ERROR ⨷ %s regularizer is not defined or supported.' % regularizer)

        return var

    def conv(self, **kwargs):
        message = kwargs.get('inputs', '_INPUTS')
        return Pod(partial(self._conv, **kwargs), ['inputs'], kwargs['name'], [message],
                   {'type': 'CONV', 'cmajor': self.channel_major, 'out_chs': kwargs.get('out_chs', None),
                    'kernel': kwargs.get('kernel', None), 'stride': kwargs.get('stride', (1, 1)),
                    'auto_pad': kwargs.get('auto_pad', True)})
    def _conv(self, name, inputs, out_chs, kernel, w_init='xavier', w_param=None, b_init=None, b_param=None,
                w_reg='l2', b_reg='l2', stride=(1, 1), dilation=(1, 1), group=1, auto_pad=True):
        dim = len(kernel)
        if dim == 1:
            conv_func = nn.Conv1D
            if self.channel_major:
                dform = 'NCW'
            else:
                dform = 'NWC'
        elif dim == 2:
            conv_func = nn.Conv2D
            if self.channel_major:
                dform = 'NCHW'
            else:
                dform = 'NHWC'
        elif dim == 3:
            conv_func = nn.Conv3D
            if self.channel_major:
                dform = 'NCDHW'
            else:
                dform = 'NDHWC'
        else:
            raise Exception('NEBULAE ERROR ⨷ %d-d convolution is not supported.' % dim)

        if self.channel_major:
            size = inputs[2:]
        else:
            size = inputs[1:-1]
        if auto_pad:
            padding = []
            for d in range(dim):
                padding.append(ceil(((ceil(size[d] / stride[d]) - 1) * stride[d] + kernel[d] + (dilation[d] - 1) * (kernel[d] - 1) - size[d]) / 2))
        else:
            padding = dim * [0]
        if not w_init is None:
            w_init = self._getInitializer(w_init, w_param, w_reg)
        if not b_init is None:
            use_bias = True
            b_init = self._getInitializer(b_init, b_param, b_reg)
        else:
            use_bias = False
        return conv_func(channels=out_chs, kernel_size=kernel, strides=stride, padding=padding, dilation=dilation,
                         groups=group, layout=dform, use_bias=use_bias, weight_initializer=w_init, bias_initializer=b_init)

    def trans_conv(self, **kwargs):
        message = kwargs.get('inputs', '_INPUTS')
        return Pod(partial(self._trans_conv, **kwargs), ['inputs'], kwargs['name'], [message],
                   {'type': 'CONV', 'cmajor': self.channel_major, 'out_chs': kwargs.get('out_chs', None),
                    'kernel': kwargs.get('kernel', None), 'stride': kwargs.get('stride', (1, 1)),
                    'auto_pad': kwargs.get('auto_pad', True)})
    def _trans_conv(self, name, inputs, output, kernel, w_init='xavier', w_param=None, b_init=None, b_param=None,
                w_reg='l2', b_reg='l2', stride=(1, 1), dilation=(1, 1), group=1, auto_pad=True):
        dim = len(kernel)
        if dim == 1:
            conv_func = nn.Conv1DTranspose
            if self.channel_major:
                dform = 'NCW'
            else:
                dform = 'NWC'
        elif dim == 2:
            conv_func = nn.Conv2DTranspose
            if self.channel_major:
                dform = 'NCHW'
            else:
                dform = 'NHWC'
        elif dim == 3:
            conv_func = nn.Conv3DTranspose
            if self.channel_major:
                dform = 'NCDHW'
            else:
                dform = 'NDHWC'
        else:
            raise Exception('NEBULAE ERROR ⨷ %d-d convolution is not supported.' % dim)

        if self.channel_major:
            in_size = inputs[2:]
            in_chs = inputs[1]
            out_chs = output[1]
            out_size = output[2:]
        else:
            in_size = inputs[1:-1]
            in_chs = inputs[-1]
            out_chs = output[-1]
            out_size = output[1:-1]
        if auto_pad:
            padding = []
            for d in range(dim):
                padding.append(ceil(((ceil(out_size[d] / stride[d]) - 1) * stride[d] + kernel[d] + (dilation[d] - 1)
                                     * (kernel[d] - 1) - out_size[d]) / 2))
        else:
            padding = dim * [0]
        compensation = []
        for d in range(dim):
            compensation.append(out_size[d] + 2 * padding[d] - kernel[d] - (in_size[d] - 1) * stride[d])
        if not w_init is None:
            w_init = self._getInitializer(w_init, w_param, w_reg)
        if not b_init is None:
            use_bias = True
            b_init = self._getInitializer(b_init, b_param, b_reg)
        else:
            use_bias = False
        return conv_func(out_chs, kernel, stride, padding, output_padding=compensation,
                         dilation=dilation, groups=group, layout=dform, use_bias=use_bias,
                         weight_initializer=w_init, bias_initializer=b_init, in_channels=in_chs)

    # ------------------------------------ Activation ------------------------------------ #

    def sigmoid(self, **kwargs):
        message = kwargs.get('inputs', '_INPUTS')
        return Pod(partial(self._sigmoid, **kwargs), ['inputs'], kwargs['name'], [message], {'type': 'ACTIVATION'})
    def _sigmoid(self, name, inputs):
        return MXsigmoid()

    def tanh(self, **kwargs):
        message = kwargs.get('inputs', '_INPUTS')
        return Pod(partial(self._tanh, **kwargs), ['inputs'], kwargs['name'], [message], {'type': 'ACTIVATION'})
    def _tanh(self, name, inputs):
        return MXtanh()

    def softmax(self, **kwargs):
        message = kwargs.get('inputs', '_INPUTS')
        return Pod(partial(self._tanh, **kwargs), ['inputs'], kwargs['name'], [message], {'type': 'ACTIVATION'})
    def _softmax(self, name, inputs):
        return MXsoftmax()

    def relu(self, **kwargs):
        message = kwargs.get('inputs', '_INPUTS')
        return Pod(partial(self._relu, **kwargs), ['inputs'], kwargs['name'], [message], {'type': 'ACTIVATION'})
    def _relu(self, name, inputs):
        return MXrelu()

    def relu_leaky(self, **kwargs):
        message = kwargs.get('inputs', '_INPUTS')
        return Pod(partial(self._relu_leaky, **kwargs), ['inputs'], kwargs['name'], [message], {'type': 'ACTIVATION'})
    def _relu_leaky(self, name, inputs, alpha=0.2):
        return nn.LeakyReLU(alpha)

    # ------------------------------------ Distributing ------------------------------------ #

    def dropout(self, **kwargs):
        message = []
        message.append(kwargs.get('inputs', '_INPUTS'))
        message.append(kwargs.get('is_train', '_IS_TRAIN'))
        return Pod(partial(self._dropout, **kwargs), ['inputs', 'is_train'], kwargs['name'], message, {'type': 'DROP'})
    def _dropout(self, name, inputs, p_drop, is_train):
        return nn.Dropout(p_drop)

    def batch_norm(self, **kwargs):
        message = []
        message.append(kwargs.get('inputs', '_INPUTS'))
        message.append(kwargs.get('is_train', '_IS_TRAIN'))
        return Pod(partial(self._batch_norm, **kwargs), ['inputs', 'is_train'], kwargs['name'], message,
                   {'type': 'BN', 'cmajor': self.channel_major})
    def _batch_norm(self, name, inputs, is_train, mmnt=0.9, beta=False, gamma=False):
        if self.channel_major:
            axis = 1
        else:
            axis = -1
        return nn.BatchNorm(axis=axis, momentum=mmnt, epsilon=1e-5, center=beta, scale=gamma)

    # def embedding(self, **kwargs):
    #     message = kwargs.get('inputs', '_INPUTS')
    #     return Pod(partial(self._embedding, **kwargs), ['inputs'], kwargs['name'], [message])
    # def _embedding(self, name, inputs, vocabulary, vec_dims, w_init='xavier', w_param=None):
    #     embd_vec = self._createVar(name+'_vec', [vocabulary, vec_dims], w_init, w_param)
    #     return tf.nn.embedding_lookup(embd_vec, inputs, name=name)

    # ------------------------------------ Pooling ------------------------------------ #

    def max_pool(self, **kwargs):
        message = kwargs.get('inputs', '_INPUTS')
        return Pod(partial(self._max_pool, **kwargs), ['inputs'], kwargs['name'], [message],
                   {'type': 'POOL', 'cmajor': self.channel_major, 'kernel': kwargs.get('kernel', (2, 2)),
                    'stride': kwargs.get('stride', (2, 2)), 'auto_pad': kwargs.get('auto_pad', True),
                    'if_global': kwargs.get('if_global', False)})
    def _max_pool(self, name, inputs, kernel=(2, 2), stride=(2, 2), auto_pad=True, if_global=False):
        dim = len(kernel)
        if dim == 1:
            pool_func = nn.MaxPool1D
            if self.channel_major:
                dform = 'NCW'
            else:
                dform = 'NWC'
        elif dim == 2:
            pool_func = nn.MaxPool2D
            if self.channel_major:
                dform = 'NCHW'
            else:
                dform = 'NHWC'
        elif dim == 3:
            pool_func = nn.MaxPool3D
            if self.channel_major:
                dform = 'NCDHW'
            else:
                dform = 'NDHWC'
        else:
            raise Exception('NEBULAE ERROR ⨷ %d-d convolution is not supported.' % dim)

        if self.channel_major:
            size = inputs[2:]
        else:
            size = inputs[1:-1]
        if if_global:
            kernel = size
            if isinstance(kernel, int):
                kernel = (kernel,)
            padding = dim * [0]
            stride = dim * [1]
        else:
            if auto_pad:
                padding = []
                for d in range(dim):
                    padding.append(ceil(((ceil(size[d] / stride[d]) - 1) * stride[d] + kernel[d] - size[d]) / 2))
            else:
                padding = dim * [0]
        return pool_func(pool_size=kernel, strides=stride, padding=padding, ceil_mode=True, layout=dform)

    def avg_pool(self, **kwargs):
        message = kwargs.get('inputs', '_INPUTS')
        return Pod(partial(self._avg_pool, **kwargs), ['inputs'], kwargs['name'], [message],
                   {'type': 'POOL', 'cmajor': self.channel_major, 'kernel': kwargs.get('kernel', (2, 2)),
                    'stride': kwargs.get('stride', (2, 2)), 'auto_pad': kwargs.get('auto_pad', True),
                    'if_global': kwargs.get('if_global', False)})
    def _avg_pool(self, name, inputs, kernel=(2, 2), stride=(2, 2), auto_pad=True, if_global=False):
        dim = len(kernel)
        if dim == 1:
            pool_func = nn.AvgPool1D
            if self.channel_major:
                dform = 'NCW'
            else:
                dform = 'NWC'
        elif dim == 2:
            pool_func = nn.AvgPool2D
            if self.channel_major:
                dform = 'NCHW'
            else:
                dform = 'NHWC'
        elif dim == 3:
            pool_func = nn.AvgPool3D
            if self.channel_major:
                dform = 'NCDHW'
            else:
                dform = 'NDHWC'
        else:
            raise Exception('NEBULAE ERROR ⨷ %d-d convolution is not supported.' % dim)

        if self.channel_major:
            size = inputs[2:]
        else:
            size = inputs[1:-1]
        if if_global:
            kernel = size
            padding = dim * [0]
            stride = dim * [1]
        else:
            if auto_pad:
                padding = []
                for d in range(dim):
                    padding.append(ceil(((ceil(size[d] / stride[d]) - 1) * stride[d] + kernel[d] - size[d]) / 2))
            else:
                padding = dim * [0]
        return pool_func(pool_size=kernel, strides=stride, padding=padding, ceil_mode=True, layout=dform)

    # ------------------------------------ Reshape ------------------------------------ #

    def reshape(self, **kwargs):
        message = kwargs.get('inputs', '_INPUTS')
        return Pod(partial(self._reshape, **kwargs), ['inputs'], kwargs['name'], [message],
                   {'type': 'RESHAPE', 'shape': kwargs.get('shape', None)})
    def _reshape(self, name, inputs, shape):
        return MXreshape(*shape)
    #
    def flat(self, **kwargs):
        message = kwargs.get('inputs', '_INPUTS')
        return Pod(partial(self._flat, **kwargs), ['inputs'], kwargs['name'], [message], {'type': 'FLAT'})
    def _flat(self, name, inputs):
        return nn.Flatten()
    #
    # def slice(self, **kwargs):
    #     message = kwargs.get('inputs', '_INPUTS')
    #     return Pod(partial(self._slice, **kwargs), ['inputs'], kwargs['name'], [message])
    # def _slice(self, name, inputs, indices):
    #     for d in range(len(indices)):
    #         output = inputs[indices[d][0]:indices[d][1]]
    #     return tf.identity(output, name)
    #
    # def pad(self, **kwargs):
    #     message = kwargs.get('inputs', '_INPUTS')
    #     return Pod(partial(self._pad, **kwargs), ['inputs'], kwargs['name'], [message])
    # def _pad(self, name, inputs, margin, fill_in=0):
    #     return tf.pad(inputs, margin, constant_values=fill_in, name=name)
    #
    # # -------------------------------------- Arithmetic -------------------------------------- #
    #
    # def clip(self, **kwargs):
    #     message = kwargs.get('inputs', '_INPUTS')
    #     return Pod(partial(self._clip, **kwargs), ['inputs'], kwargs['name'], [message])
    # def _clip(self, name, inputs, min_max_vals):
    #     return tf.clip_by_value(inputs, min_max_vals[0], min_max_vals[1], name)

    def dense(self, **kwargs):
        message = kwargs.get('inputs', '_INPUTS')
        return Pod(partial(self._dense, **kwargs), ['inputs'], kwargs['name'], [message],
                   {'type': 'DENSE', 'out_chs': kwargs.get('out_chs', None), 'axis': kwargs.get('axis', -1)})
    def _dense(self, name, inputs, out_chs, axis=-1, w_init='xavier', w_param=None,
               b_init='zero', b_param=None, w_reg='l2', b_reg='l2'):
        if axis == 0:
            raise Exception('NEBULAE ERROR ⨷ you cannot apply dense layer along batch axis.')
        elif axis == -1:
            in_shape = None
            out_shape = None
        else:
            size = list(inputs)
            channel = (size[axis],)
            del size[axis]
            size[0] = -1
            in_shape = tuple(size) + channel
            out_shape = list(inputs)
            out_shape[0] = -1
            out_shape[axis] = out_chs
            out_shape = tuple(out_shape)

        if not w_init is None:
            w_init = self._getInitializer(w_init, w_param, w_reg)
        if not b_init is None:
            use_bias = True
            b_init = self._getInitializer(b_init, b_param, b_reg)
        else:
            use_bias = False
        return MXdense(out_chs, use_bias, w_init, b_init, in_shape, out_shape)

    # def argmax(self, **kwargs):
    #     message = kwargs.get('inputs', '_INPUTS')
    #     return Pod(partial(self._argmax, **kwargs), ['inputs'], kwargs['name'], [message])
    # def _argmax(self, name, inputs, axis):
    #     return tf.argmax(inputs, axis, name=name)
    #

    # ---------------------------------------- Wrapped Net ---------------------------------------- #



    # # ------------------------------------ Image Manipulation ------------------------------------ #
    #
    # def resize(self, **kwargs):
    #     message = kwargs.get('inputs', '_INPUTS')
    #     return Pod(partial(self._resize, **kwargs), ['inputs'], kwargs['name'], [message])
    # def _resize(self, name, inputs, size, method='bilinear'):
    #     if method == 'bilinear':
    #         return tf.image.resize_bilinear(inputs, size, name=name)
    #     elif method == 'bicubic':
    #         return tf.image.resize_bicubic(inputs, size, name=name)
    #     elif method == 'crop':
    #         return tf.image.resize_image_with_crop_or_pad(inputs, size, name=name)
    #     else:
    #         raise KeyError('NEBULAE ERROR ⨷ %s is not a legal resize method.' % method)
    #
    #
    # # ------------------------------------ Redefine or Rename ------------------------------------ #
    #
    def duplicate(self, **kwargs):
        message = kwargs.get('inputs', '_INPUTS')
        return Pod(partial(self._duplicate, **kwargs), ['inputs'], kwargs['name'], [message], {'type': 'DUP'})
    def _duplicate(self, name, inputs):
        return Duplicate()
    #
    # def convert(self, **kwargs):
    #     message = kwargs.get('inputs', '_INPUTS')
    #     return Pod(partial(self._convert, **kwargs), ['inputs'], kwargs['name'], [message])
    # def _convert(self, name, inputs, dtype, trainable=False):
    #     '''
    #     convert data type or convert list/numpy array to tensor
    #     :param name:
    #     :param inputs: inputs tensor / list / numpy array
    #     :param dtype:
    #     :param trainable: if tensor is trainable
    #     :return: tensor
    #     '''
    #     if isinstance(inputs, (tf.Tensor, tf.SparseTensor, tf.Variable)):
    #         return tf.cast(inputs, tf.as_dtype(dtype), name=name)
    #     else:
    #         return tf.Variable(inputs, trainable=trainable, name=name)

    # ------------------------------------ Loss ------------------------------------ #

    def weight_decay(self, **kwargs):
        return Pod(partial(self._weight_decay, **kwargs), ['penalty'], kwargs['name'],
                   [nd.ones((1)) * kwargs['penalty']], {'type': 'LOSS'})
    def _weight_decay(self, name, penalty, decay_scope=None):
        if not decay_scope is None:
            raise Exception('NEBULAE ERROR ⨷ weight decay cannot be manipulated explicitly in MXNet core.')
        return MXwd()

    def sigm_xentropy(self, **kwargs):
        message = []
        message.append(kwargs.get('inputs', '_INPUTS'))
        message.append(kwargs.get('label', '_LABEL'))
        return Pod(partial(self._sigm_xentropy, **kwargs), ['inputs', 'label'], kwargs['name'], message, {'type': 'LOSS'})
    def _sigm_xentropy(self, name, inputs, label):
        return MXsigmxe()

    def sftm_xentropy(self, **kwargs):
        message = []
        message.append(kwargs.get('inputs', '_INPUTS'))
        message.append(kwargs.get('label', '_LABEL'))
        return Pod(partial(self._sftm_xentropy, **kwargs), ['inputs', 'label'], kwargs['name'], message, {'type': 'LOSS'})
    def _sftm_xentropy(self, name, inputs, label, is_one_hot):
        return MXsftmxe(is_one_hot)

    def mse(self, **kwargs):
        message = []
        message.append(kwargs.get('inputs', '_INPUTS'))
        message.append(kwargs.get('label', '_LABEL'))
        return Pod(partial(self._mse, **kwargs), ['inputs', 'label'], kwargs['name'], message, {'type': 'LOSS'})
    def _mse(self, name, inputs, label):
        return MXmse()

    def mae(self, **kwargs):
        message = []
        message.append(kwargs.get('inputs', '_INPUTS'))
        message.append(kwargs.get('label', '_LABEL'))
        return Pod(partial(self._mae, **kwargs), ['inputs', 'label'], kwargs['name'], message, {'type': 'LOSS'})
    def _mae(self, name, inputs, label):
        return MXmae()

    # ------------------------------------ Optimizer ------------------------------------ #

    def _lrStrategy(self, lr, lr_decay, miles, param):
        if lr_decay == 'step':
            return lr_scheduler.FactorScheduler(step=miles, factor=param, base_lr=lr)
        elif lr_decay == 'poly':
            return lr_scheduler.PolyScheduler(max_update=miles, base_lr=lr, pwr=param)
        elif lr_decay == 'cosine':
            return lr_scheduler.CosineScheduler(miles, base_lr=lr, final_lr=0.001*lr)
        elif lr_decay == 'exp':
            pass  #return tf.train.piecewise_constant(mileage, [i*param for i in range(1, len(lr))], lr)
        elif lr_decay == 'wavy':
            return CyclicalSchedule(CosineAnnealingSchedule, min_lr=0.001*lr, max_lr=lr,
                             cycle_length=miles, cycle_length_decay=1, cycle_magnitude_decay=param)
        else:
            raise KeyError('NEBULAE ERROR ⨷ %s decay is not supported or defined.' % lr_decay)

    def _selectParam(self, global_var, update_scope, ignore_name):
        update_var = ParameterDict()
        if ignore_name is None:
            ignore_name = []
        for l, p in global_var.items():
            if (update_scope is None or l.startswith(update_scope)) and (l not in ignore_name):
                update_var.update(p)
        return update_var

    def momentum(self, **kwargs):
        recargs = {k: v for k, v in kwargs.items() if k not in ('name', 'inputs', 'media')}
        recordConfig(os.path.join(os.getcwd(), 'nebulae_temp_config.json'),
                     recargs, overwrite=False)

        message = kwargs.get('inputs', '_INPUTS')
        return Pod(partial(self._momentum, **kwargs), ['inputs', 'media'], kwargs['name'], [message, '_MEDIA'], {'type': 'OPTZ'})
    def _momentum(self, name, lr, inputs=None, mmnt=0.9, update_scope=None, ignore_name=None,
                  lr_decay=None, lr_miles=None, decay_param=None, grad_limit=None, media=None):
        # N.B. media is network and wd factor
        mod_params, wd = media
        wd = wd.asscalar()
        if isinstance(lr_decay, str):
            lr_s = self._lrStrategy(lr, lr_decay, lr_miles, decay_param)
        else:
            lr_s = lr_scheduler.LRScheduler(base_lr=lr)
        update_params = self._selectParam(mod_params, update_scope, ignore_name)
        optz = Trainer(update_params, 'SGD',
                       {'learning_rate': lr, 'momentum': mmnt, 'wd': wd,
                        'lr_scheduler': lr_s, 'clip_gradient': grad_limit})
        return optz

    def nesterov(self, **kwargs):
        recargs = {k: v for k, v in kwargs.items() if k not in ('name', 'inputs', 'media')}
        recordConfig(os.path.join(os.getcwd(), 'nebulae_temp_config.json'),
                     recargs, overwrite=False)

        message = kwargs.get('inputs', '_INPUTS')
        return Pod(partial(self._nesterov, **kwargs), ['inputs', 'media'], kwargs['name'], [message, '_MEDIA'], {'type': 'OPTZ'})
    def _nesterov(self, name, lr, inputs=None, mmnt=0.9, update_scope=None, ignore_name=None,
                  lr_decay=None, lr_miles=None, decay_param=None, grad_limit=None, media=None):
        # N.B. media is network and wd factor
        mod_params, wd = media
        wd = wd.asscalar()
        if isinstance(lr_decay, str):
            lr_s = self._lrStrategy(lr, lr_decay, lr_miles, decay_param)
        else:
            lr_s = lr_scheduler.LRScheduler(base_lr=lr)
        update_params = self._selectParam(mod_params, update_scope, ignore_name)
        optz = Trainer(update_params, 'NAG',
                       {'learning_rate': lr, 'momentum': mmnt, 'wd': wd,
                        'lr_scheduler': lr_s, 'clip_gradient': grad_limit})
        return optz

    def adam(self, **kwargs):
        recargs = {k: v for k, v in kwargs.items() if k not in ('name', 'inputs', 'media')}
        recordConfig(os.path.join(os.getcwd(), 'nebulae_temp_config.json'),
                     recargs, overwrite=False)

        message = kwargs.get('inputs', '_INPUTS')
        return Pod(partial(self._adam, **kwargs), ['inputs', 'media'], kwargs['name'], [message, '_MEDIA'], {'type': 'OPTZ'})
    def _adam(self, name, lr, inputs=None, mmnt1=0.9, mmnt2=0.999, update_scope=None, ignore_name=None,
                  lr_decay=None, lr_miles=None, decay_param=None, grad_limit=None, media=None):
        # N.B. media is network and wd factor
        mod_params, wd = media
        wd = wd.asscalar()
        if isinstance(lr_decay, str):
            lr_s = self._lrStrategy(lr, lr_decay, lr_miles, decay_param)
        else:
            lr_s = lr_scheduler.LRScheduler(base_lr=lr)
        update_params = self._selectParam(mod_params, update_scope, ignore_name)
        optz = Trainer(update_params, 'Adam',
                       {'learning_rate': lr, 'beta1': mmnt1, 'beta2': mmnt2, 'wd': wd,
                        'lr_scheduler': lr_s, 'clip_gradient': grad_limit})
        return optz

    # ------------------------------------ Metric ------------------------------------ #

    def accuracy_cls(self, **kwargs):
        message = []
        message.append(kwargs.get('inputs', '_INPUTS'))
        message.append(kwargs.get('label', '_LABEL'))
        return Pod(partial(self._accuracy_cls, **kwargs), ['inputs', 'label'], kwargs['name'], message, {'type': 'ACC'})
    def _accuracy_cls(self, name, inputs, label, multi_class=False, is_one_hot=False):
        return AccuracyCLS(multi_class, is_one_hot)


#!/usr/bin/env python
'''
time_machine_tf
Created by Seria at 04/02/2019 4:35 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-
import torch
import os
from glob import glob



class TimeMachinePT(object):
    def __init__(self, param):
        '''
        Time Machine saves current states or restores saved states
        '''
        self.param = param
        self.counter = 1

    def to(self, craft, file='', ckpt_scope=None, frozen=False, ins=None, outs=None):
        assert self.param['ckpt_path'] is not None, Exception('NEBULAE ERROR ⨷ anchor location is not provided.')

        ckpt_path = os.path.join(self.param['ckpt_path'], file)
        if os.path.isfile(ckpt_path):
            moment = ckpt_path
        else:
            architecture = glob(os.path.join(ckpt_path,'*.pth'))
            max_saved = -1
            moment = None
            for arch in architecture:
                saved_no = int(arch.split('.')[0].split('-')[-1])
                if saved_no > max_saved:
                    moment = arch
                    max_saved = saved_no
        if moment is None:
            raise Exception('NEBULAE ERROR ⨷ valid anchor is not found.')

        states = torch.load(moment)
        if not ckpt_scope is None:
            ckpt_scope = ckpt_scope.replace('/', '.')
            states = {k: v for k, v in states.items() if k.startswith(ckpt_scope)}
        craft.load_state_dict(states, strict=frozen)
        print('+' + ((10 + len(moment)) * '-') + '+')
        print('| Back to \033[1;34m%s\033[0m |' % moment)
        print('+' + ((10 + len(moment)) * '-') + '+')

    def drop(self, craft, file='', save_scope=None, frozen=False, anchor=None):
        if self.param['rank']>0:
            return
        assert self.param['save_path'] is not None, Exception('NEBULAE ERROR ⨷ there is nowhere to drop anchor.')

        save_path = os.path.join(self.param['save_path'], file)
        states = craft.state_dict()
        if save_scope is not None:
            save_scope = save_scope.replace('/', '.')
            states = {k:v for k,v in states.items() if k.startswith(save_scope)}
        if save_path.endswith('.pth'):
            save_ckpt = save_path
        else:
            save_ckpt = os.path.join(save_path, '%s-%d.pth'%(craft.scope, self.counter))
        torch.save(states, save_ckpt)
        self.counter += 1
        print('| Anchor is dropped at \033[1;34m%s\033[0m |' % save_ckpt)


#!/usr/bin/env python
'''
garage
Created by Seria at 03/01/2019 8:32 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-
from ... import dock
from .architect import ResG, ResD, BN


class Discriminator(dock.Craft):
    def __init__(self, in_shape, base_chs, norm_fn, attention, spec_norm, w_init, scope='DSC'):
        super(Discriminator, self).__init__(scope)
        H, W, C = in_shape
        min_size = min(H, W)
        factor = {128: base_chs/16}
        self.backbone = ResD(in_shape, base_chs, norm_fn, attention, spec_norm, w_init)
        if spec_norm:
            self.cls = dock.SN(dock.Dense(int(H * W * factor[min_size]), 1))
        else:
            self.cls = dock.Dense(int(H * W * factor[min_size]), 1)

    def run(self, x):
        x = self.backbone(x)
        self['out'] = self.cls(x)

        return self['out']



class ResGAN(dock.Craft):
    def __init__(self, in_shape, latent_dim=128, base_chs=32, norm_fn=BN, attention=False,
                 spec_norm=False, w_init=dock.XavierNorm(), scope='RESGAN'):
        super(ResGAN, self).__init__(scope)
        self.G = ResG(in_shape, latent_dim, base_chs, norm_fn, attention, spec_norm, w_init)
        self.D = Discriminator(in_shape, base_chs, norm_fn, attention, spec_norm, w_init)

    def run(self):
        pass


#!/usr/bin/env python
'''
engine_mx
Created by Seria at 12/02/2019 3:45 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========

'''
# -*- coding:utf-8 -*-
import torch
import os
import multiprocessing as mp
# from apex import parallel
from types import MethodType


class DDP(torch.nn.parallel.DistributedDataParallel):
    #     def __init__(self, module, delay_allreduce):
    def __init__(self, module, device_ids, output_device):
        #         super(DDP, self).__init__(module, delay_allreduce=delay_allreduce)
        super(DDP, self).__init__(module, device_ids=device_ids, output_device=output_device)

    def __getattr__(self, name: str):
        if '_parameters' in self.__dict__:
            _parameters = self.__dict__['_parameters']
            if name in _parameters:
                return _parameters[name]
        if '_buffers' in self.__dict__:
            _buffers = self.__dict__['_buffers']
            if name in _buffers:
                return _buffers[name]
        if '_modules' in self.__dict__:
            modules = self.__dict__['_modules']
            if name in modules:
                return modules[name]
        if hasattr(self, 'module'):
            if hasattr(self.module, name):
                return getattr(self.module, name)
        raise AttributeError("'{}' object has no attribute '{}'".format(type(self).__name__, name))


# DDP = parallel.DistributedDataParallel

class Multiverse(object):
    '''
    Args:
    nworld: world size
    '''

    def __init__(self, universe, nworld=1):
        self.universe = universe
        self.nworld = nworld
        self.rank = -1
        self.env = os.environ.copy()
        self.env["MASTER_ADDR"] = '127.0.0.1'
        self.env["MASTER_PORT"] = '29500'
        self.env["WORLD_SIZE"] = str(nworld)
        self.env["OMP_NUM_THREADS"] = '1'

    def __call__(self):
        # mp.set_start_method('spawn')
        ps = []
        for r in range(self.nworld):
            self.rank = r
            self.env['RANK'] = str(r)
            self.env['LOCAL_RANK'] = str(r)
            p = mp.Process(target=self.universe, args=(self,))
            p.start()
            ps.append(p)
        for p in ps:
            p.join()

    def init(self):
        for k, v in self.env.items():
            os.environ[k] = v

    def setup(self):
        torch.backends.cudnn.benchmark = True
        torch.cuda.set_device(self.rank)
        torch.distributed.init_process_group(backend='nccl', init_method='env://')

    def collect(self, model):
        scope = model.scope
        #         model = parallel.convert_syncbn_model(model)
        model = torch.nn.SyncBatchNorm.convert_sync_batchnorm(model)
        model = model.to(torch.device('cuda:%d' % self.rank))
        #         model = DDP(model, delay_allreduce=True)
        model = DDP(model, device_ids=[self.rank], output_device=self.rank)

        def gear(self, gr):
            if isinstance(gr, bool):
                if gr:
                    self.train()
                else:
                    self.eval()
            else:
                raise Exception('NEBULAE ERROR ⨷ %s is not a valid type of collected gear.' % type(gr))

        model.gear = MethodType(gear, model)
        model.scope = scope
        return model

    def reduce(self, tensor):
        rt = tensor.clone()
        torch.distributed.all_reduce(rt, op=torch.distributed.ReduceOp.SUM)
        rt /= self.nworld
        return rt


#!/usr/bin/env python
'''
engine_tf
Created by Seria at 04/02/2019 4:31 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-
import tensorflow as tf
from ..toolkit.utility import getAvailabelGPU

class EngineTF(object):
    '''
    Param:
    device: 'gpu' or 'cpu'
    available_gpus
    gpu_mem_fraction
    if_conserve
    least_mem
    '''
    def __init__(self, param):
        self.param = param
        # look for available gpu devices
        self.config_proto = tf.ConfigProto(log_device_placement=False)
        self.config_proto.gpu_options.allow_growth = self.param['if_conserve']
        if self.param['device'].lower() == 'gpu':
            self.config_proto.gpu_options.per_process_gpu_memory_fraction = self.param['gpu_mem_fraction']
            if len(self.param['available_gpus'])==0:
                gpus = getAvailabelGPU(self.param['ngpus'], self.param['least_mem'])
                if len(gpus) == 0:
                    raise Exception('NEBULAE ERROR ⨷ no enough available gpu', gpus)
                # TODO: multi-gpu training is to be supported
                gpus = str(gpus[0])
            else:
                gpustr = ''
                for ag in self.param['available_gpus']:
                    gpustr += str(ag)+','
                gpus = gpustr[:-1]
            self.device = 'gpu:' + gpus
            self.config_proto.gpu_options.visible_device_list = gpus
            print('+' + ((24 + len(gpus)) * '-') + '+')
            print('| Reside in Device: \033[1;36mGPU-%s\033[0m |' % gpus)
            print('+' + ((24 + len(gpus)) * '-') + '+')
        elif self.param['device'].lower() == 'cpu':
            self.device = 'cpu'
            print('+' + (23 * '-') + '+')
            print('| Reside in Device: \033[1;36mCPU\033[0m |')
            print('+' + (23 * '-') + '+')
        else:
            raise KeyError('NEBULAE ERROR ⨷ given device should be either cpu or gpu.')


#!/usr/bin/env python
'''
utility
Created by Seria at 14/11/2018 8:33 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-

import json
import numpy as np
import subprocess as subp
import h5py
import os
import io
from math import ceil
from PIL import Image
from copy import deepcopy
from ..law import Constant



def autoPad(in_size: tuple, kernel: tuple, stride=1, dilation=1):
    '''
    Args
    - in_size: input size e.g. (h, w) for 2d tensor
    - kernel: kernel size
    - stride: convolution stride
    - dilation: stride in atrous convolution

    Return
    padding elements on dimensions in reverse order
    e.g. (left, right, top, bottom, front, back) for 3d tensor
    '''
    dim = len(kernel)
    if isinstance(stride, int):
        stride = dim * [stride]
    if isinstance(dilation, int):
        dilation = dim * [dilation]

    padding = []
    for d in range(dim-1,-1,-1):
        margin = (ceil(in_size[d] / stride[d]) - 1) * stride[d] + kernel[d] + (dilation[d] - 1) \
                             * (kernel[d] - 1) - in_size[d]
        padding.extend([margin//2, margin-margin//2])
    return padding

def cap(key, scope):
    return scope + '/' + key

def doff(key):
    return key.split('/')[-1]

def leftOnExceptHook(exc_type, exc_value, tb, ignored=''):
    '''
    leave particular exception messages out
    '''
    msg = ' Traceback (most recent call last):\n'
    while tb:
        filename = tb.tb_frame.f_code.co_filename
        name = tb.tb_frame.f_code.co_name
        lineno = tb.tb_lineno
        if not ignored or ignored not in filename:
            msg += '   File "%.500s", line %d, in %.500s\n' % (filename, lineno, name)
        tb = tb.tb_next

    msg += ' %s: %s\n' %(exc_type.__name__, exc_value)
    print(msg)


def toDenseLabel(labels, nclasses, on_value=1, off_value=0):
    batch_size = labels.shape[0]
    # initialize dense labels
    dense = off_value * np.ones((batch_size * nclasses), dtype=np.float32)
    indices = []
    if isinstance(labels[0], str):
        for b in range(batch_size):
            indices += [int(s) + b * nclasses for s in labels[b].split(Constant.CHAR_SEP)]
    elif isinstance(labels[0], (list, np.ndarray)): # labels is a nested array
        for b in range(batch_size):
            indices += [l + b * nclasses for l in labels[b]]
        dense[indices] = on_value
    else:
        for b in range(batch_size):
            indices += [int(labels[b]) + b * nclasses]
        dense[indices] = on_value
    return np.reshape(dense, (batch_size, nclasses))


def getAvailabelGPU(ngpus, least_mem):
    p = subp.Popen('nvidia-smi', stdout=subp.PIPE)
    gpu_id = 0  # next gpu we are about to probe
    flag_gpu = False
    id_mem = []  # gpu having avialable memory greater than least requirement
    for l in p.stdout.readlines():
        line = l.decode('utf-8').split()
        if len(line) < 1:
            break
        elif len(line) < 2:
            continue
        if line[1] == str(gpu_id):
            flag_gpu = True
            continue
        if flag_gpu:
            vacancy = int(line[10].split('M')[0]) - int(line[8].split('M')[0])
            if vacancy > least_mem:
                if len(id_mem) < ngpus:
                    id_mem.append((gpu_id, vacancy)) # (id, mem) of gpu
                else:
                    id_mem.sort(key=lambda x: x[1])
                    if vacancy > id_mem[0][1]:
                        id_mem[0] = (gpu_id, vacancy)
            gpu_id += 1
            flag_gpu = False
    return [elem[0] for elem in id_mem]

def parseConfig(config_path):
    with open(config_path, 'r') as config_file:
        return json.load(config_file)

def recordConfig(config_path, config, overwrite=True):
    while not overwrite and os.path.exists(config_path):
        config_path = config_path[:-5]+'_.json'
    with open(config_path, 'w') as config_file:
        json.dump(config, config_file, indent=4)

def _mergeFuel(src_dir, src, dst, dtype, keep_src=True):
    data = {}
    info_keys = []
    shards = len(src)
    print('+' + (23 + 2 * shards + len(dst)) * '-' + '+')
    # read
    ending_char = '\r'
    for i, f in enumerate(src):
        hdf5 = h5py.File(os.path.join(src_dir, f), 'r')
        if i == 0:
            for key in hdf5.keys():
                info_keys.append(key)
            for key in info_keys:
                if key == Constant.FRAME_KEY:
                    data[key] = 0
                else:
                    data[key] = []
        for key in info_keys:
            if key == Constant.FRAME_KEY:
                frames = hdf5[key]
                data[key] = frames if frames>data[key] else data[key]
            else:
                data[key].extend(hdf5[key][()].tolist())
        hdf5.close()
        if not keep_src:
            os.remove(os.path.join(src_dir, f))

        progress = i+1
        yellow_bar = progress * '❖-'
        space_bar = (shards - progress) * '◇-'
        if progress == shards:
            ending_char = '\n'
        print('| Merging data  \033[1;35m%s\033[0m  ⊰⟦-%s%s⟧⊱ |'
              % (dst, yellow_bar, space_bar), end=ending_char)
    # write
    hdf5 = h5py.File(os.path.join(src_dir, dst), 'w')
    for key in info_keys:
        if dtype[key].startswith('v'):  # dealing with encoded / varied data
            dt = h5py.special_dtype(vlen=dtype[key])
            hdf5.create_dataset(key, dtype=dt, data=np.array(data[key]))
        elif key == Constant.FRAME_KEY:
            hdf5[key] = data[key]
        else:
            hdf5[key] = np.array(data[key]).astype(dtype[key])
    hdf5.close()
    print('+' + (23 + 2 * shards + len(dst)) * '-' + '+')

def _fillFuel(src, key, data):
    hdf5 = h5py.File(src, 'a')
    if isinstance(data, (int, float, str)):
        data = [data]
    if isinstance(data, list):
        data = np.array(data)
    if data.dtype.kind == 'U': # convert unicode to string
        nbyte = data.dtype.descr[0][1].split('U')[-1]
        data_copy = data.copy()
        data = np.empty(data_copy.shape, dtype='|S'+nbyte)
        for idx, elm in np.ndenumerate(data_copy):
            data[idx] = elm.encode()
    if data.dtype.kind == 'S':
        sdt = h5py.special_dtype(vlen=str)
        hdf5.create_dataset(key, dtype=sdt, data=data)
    else:
        hdf5[key] = data
    hdf5.close()

def _deductFuel(src, key):
    hdf5 = h5py.File(src, 'a')
    del hdf5[key]
    hdf5.close()

def byte2arr(data_src, as_np=True):
    data_bytes = data_src.tobytes()
    data_pil = Image.open(io.BytesIO(data_bytes))
    if as_np:
        data_np = np.array(data_pil)  # pixels range between [0, 255]
        data_np = (data_np / 255).astype(np.float32)  # pixels range between [0, 1]
        return data_np
    else:
        return data_pil

def curve2str(curve, divisor, span, is_global, x_title='x', y_title='y'):
    assert curve.ndim == 1 and span > 9 # must be a vector
    assert os.get_terminal_size().columns > span + 10 + 5 # ensure that curve is not too long to display

    line_type = {'ascent': '/', 'descent': '\\', 'vertical': '|', 'horizontal': '_'}
    if span < curve.size:
        if is_global:
            indices = [round(i * curve.size / span) for i in range(span)]
            curve = curve[indices]
            indices = [idx+1 for idx in indices]
        else:
            indices = [curve.size - i for i in range(span, 0, -1)]
            curve = curve[-span:]
    else:
        indices = [i for i in range(1, span+1)]
    y_max = curve.max()
    y_min = curve.min()
    delta = (y_max - y_min) / divisor
    if delta == 0.:
        grid = [[10*' ' + ' ┃ ', span * ' ', '\n'] for i in range(1, divisor + 1)]
        grid.append([10 * ' ' + ' ▲ ', span * ' ', '\n'])
    else:
        quant = np.round((curve - y_min) / delta).astype(np.int8)

        grid = [[f'{y_min + i*delta:>10.3f} ┃ ', span * ' ', '\n'] for i in range(1, divisor + 1)]
        grid.append([10 * ' ' + ' ▲︎ ', span * ' ', '\n'])
        # draw the curve
        for i in range(curve.size-1):
            prev = quant[i]
            curr = quant[i + 1]
            if prev > curr:
                grid[prev - 1][1] = grid[prev - 1][1][:i] + line_type['descent'] + grid[prev - 1][1][i + 1:]
                for j in range(1, prev - curr):
                    grid[prev - 1 - j][1] = grid[prev - 1 - j][1][:i] + line_type['vertical'] + grid[prev - 1 - j][1][
                                                                                                i + 1:]
            elif prev < curr:
                grid[prev][1] = grid[prev][1][:i] + line_type['ascent'] + grid[prev][1][i + 1:]
                for j in range(1, curr - prev):
                    grid[prev + j][1] = grid[prev + j][1][:i] + line_type['vertical'] + grid[prev + j][1][i + 1:]
            else:
                grid[prev][1] = grid[prev][1][:i] + line_type['horizontal'] + grid[prev][1][i + 1:]

    # initialize axis
    cstr = ''
    cstr += 10 * ' ' + ' %s\n'%y_title
    for i in range(divisor, -1, -1):
        cstr += ''.join(grid[i])
    cstr += f'{y_min:>10.3f} ┗━' + span * '━' + ' ► %s\n'%x_title
    x_domain = (10+3) * ' '
    for i in range(0, len(indices), 5):
        idx = indices[i]
        if idx<1e3:
                x_domain += f'{idx:<4d} '
        elif idx<1e4:
                x_domain += f'{idx/1e3:<3.1f}K '
        elif idx<1e6:
                x_domain += f'{round(idx/1e3):<3d}K '
        elif idx<1e7:
                x_domain += f'{idx/1e6:<3.1f}M '
        elif idx<1e9:
                x_domain += f'{round(idx/1e6):<3d}M '
    cstr += x_domain+'\n'

    return cstr



def joinImg(imgs, nrow, ncol):
    N, H, W, C = imgs.shape
    assert N == nrow*ncol, 'NEBULAE ERROR ⨷ the number of images does not match cells.'
    assert N > 1, 'NEBULAE ERROR ⨷ one image does not need to be pieced together.'
    margin_h = max(1, H//20)
    margin_w = max(1, W//20)
    canvas = np.zeros((margin_h*(nrow+1) + H*nrow, margin_w*(ncol+1) + W*ncol, C))
    for c in range(ncol):
        for r in range(nrow):
            y = margin_h*(r+1) + H*r
            x = margin_w*(c+1) + W*c
            canvas[y:y+H, x:x+W] = imgs[r*ncol + c]
    return canvas


#!/usr/bin/env python
'''
garage
Created by Seria at 03/01/2019 8:32 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-
from ... import dock

from math import ceil



class Bottleneck(dock.Craft):
    def __init__(self, in_shape, neck_chs, body_chs, stride=1, dsample=False, scope='BOTTLENECK'):
        super(Bottleneck, self).__init__(scope)
        H, W, C = in_shape
        self.body_chs = body_chs
        self.conv_1 = dock.Conv(C, neck_chs, (1, 1))
        self.bn_1 = dock.BN(neck_chs, dim=2)
        pad = dock.autoPad((H, W), (3, 3), stride=stride)
        self.conv_2 = dock.Conv(neck_chs, neck_chs, (3, 3), stride=stride, padding=pad)
        self.bn_2 = dock.BN(neck_chs, dim=2)
        self.conv_3 = dock.Conv(neck_chs, self.body_chs, (1, 1))
        self.bn_3 = dock.BN(self.body_chs, dim=2)
        self.relu = dock.Relu()

        self.dsample = dsample
        if dsample:
            self.ds_conv = dock.Conv(C, self.body_chs, (1, 1), stride)
            self.ds_bn = dock.BN(self.body_chs, dim=2)

    def run(self, x):
        self['identity'] = x

        y = self.conv_1(x)
        y = self.bn_1(y)
        y = self.relu(y)

        y = self.conv_2(y)
        y = self.bn_2(y)
        y = self.relu(y)

        y = self.conv_3(y)
        y = self.bn_3(y)

        if self.dsample:
            self['identity'] = self.ds_bn(self.ds_conv(x))

        y += self['identity']
        y = self.relu(y)

        return y


class ResBlock(dock.Craft):
    def __init__(self, in_shape, nchs, nblock, stride=1, scope='RESBLOCK'):
        super(ResBlock, self).__init__(scope)
        H, W, C = in_shape
        self.nblock = nblock
        self.layer_1 = Bottleneck(in_shape, nchs, C*4, stride, dsample=True)
        for i in range(1, nblock):
            setattr(self, 'layer_%d'%i, Bottleneck((H, W, C * 4**i), nchs, C * 4**(i+1)))

    def run(self, x):
        self['input'] = x
        for i in range(1, self.nblock):
            l = getattr(self, 'layer_%d'%i)
            x = l(x)

        return x


class Resnet_V2(dock.Craft):
    def __init__(self, in_shape, nblocks, scope='RESNET_V2'):
        super(Resnet_V2, self).__init__(scope)
        H, W, C = in_shape
        self.H, self.W, self.C = H, W, C

        pad = dock.autoPad((H, W), (7, 7), 2)
        self.conv_0 = dock.Conv(C, 64, (7, 7), stride=2, padding=pad)
        self.bn_0 = dock.BN(64, dim=2)
        self.relu = dock.Relu()
        pad = dock.autoPad((ceil(H/2), ceil(W/2)), (3, 3), 2)
        self.mpool = dock.MaxPool((3, 3), padding=pad)

        self.blocks = []
        ds_power = 2
        ichs = 64
        nchs = 64
        for i, nblk in enumerate(nblocks):
            if i == 0:
                stride = 1
            else:
                stride = 2
                ds_power += 2
            self.blocks.append(ResBlock((ceil(H/2**ds_power), ceil(W/2**ds_power), ichs), nchs, nblk, stride))
            ichs = self.blocks[-1].C
            nchs *= 2

        self.gap = dock.AvgPool((-1, -1))

    def run(self, x):
        self['input'] = x
        y = self.conv_0(self['input'])
        y = self.bn_0(y)
        y = self.relu(y)
        y = self.mpool(y)

        for blk in self.blocks:
            y = blk(y)

        self['out'] = self.gap(y)
        return self['out']



class Resnet_V2_50(dock.Craft):
    def __init__(self, in_shape, scope='RESNET_V2_50'):
        super(Resnet_V2_50, self).__init__(scope)
        self.backbone = Resnet_V2(in_shape, [3, 4, 6, 3])

    def run(self, x):
        self['input'] = x
        self['out'] = self.backbone(self['input'])
        return self['out']


class Resnet_V2_101(dock.Craft):
    def __init__(self, in_shape, scope='RESNET_V2_101'):
        super(Resnet_V2_101, self).__init__(scope)
        self.backbone = Resnet_V2(in_shape, [3, 4, 23, 3])

    def run(self, x):
        self['input'] = x
        self['out'] = self.backbone(self['input'])
        return self['out']


class Resnet_V2_152(dock.Craft):
    def __init__(self, in_shape, scope='RESNET_V2_152'):
        super(Resnet_V2_152, self).__init__(scope)
        self.backbone = Resnet_V2(in_shape, [3, 8, 36, 3])

    def run(self, x):
        self['input'] = x
        self['out'] = self.backbone(self['input'])
        return self['out']